"""
Simple MongoDB Document Wrapper

Adds useful functionality beyond standard dict type.

"""

import re
import unicodedata

from pymongo.database import Database
from bson.objectid import ObjectId, InvalidId

_slugify_strip_re = re.compile(r'[^\w\s-]')
_slugify_hyphenate_re = re.compile(r'[-\s]+')


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    From Django's "django/template/defaultfilters.py".
    """
    if not isinstance(value, unicode):
        value = unicode(value)
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    value = unicode(_slugify_strip_re.sub('', value).strip().lower())
    return _slugify_hyphenate_re.sub('-', value)


class NoDatabaseSet(Exception):
    """ No database on document """
    pass


class DocumentNotSaved(Exception):
    """ Document not saved """
    pass


class Document(dict):
    """
    SimpleMongo Document/Collection wrapper

    """

    # if unset, uses class name for Mongo collection
    __collection__ = None
    __database__ = None

    def __init__(self):
        if not isinstance(self.__database__, Database):
            raise NoDatabaseSet()
        collection = self.__collection__ or self.__class__.__name__
        self.collection = self.__database__[collection]

    @classmethod
    def find(cls, *args, **kwargs):
        collection = cls.__collection__ or cls.__name__
        if not isinstance(cls.__database__, Database):
            raise NoDatabaseSet()
        return cls.__database__[collection].find(*args, as_class=cls, **kwargs)

    @classmethod
    def find_one(cls, *args, **kwargs):
        collection = cls.__collection__ or cls.__name__
        if not isinstance(cls.__database__, Database):
            raise NoDatabaseSet()
        return cls.__database__[collection].find_one(
            *args, as_class=cls, **kwargs)

    def save(self):
        return self.collection.save(self)

    def delete(self):
        if not self.get('_id'):
            raise DocumentNotSaved()
        return self.collection.remove({'_id': self['_id']})

    def slugify(self, from_fields=None, from_text=None):
        """ Generate slug from specified fields or text """
        if not from_text:
            if not from_fields and not hasattr(self, 'slug_fields'):
                raise ValueError('No implicit or explicit fields to slugify')
            in_fields = from_fields or self.slug_fields
            if type(in_fields) is list:
                from_text = ' '.join(
                    [unicode(self[data]) for data in in_fields])
            else:
                from_text = unicode(self[in_fields])
        slug = slugify(from_text)
        next_id = self.__database__['__slug_counter'].find_and_modify(
            query={'_id': slug},
            update={'$inc': {'count': 1}},
            upsert=True,
            new=True)
        if next_id.get('count') == 1:
            return slug
        return '%s-%d' % (slug, next_id.get('count'))

    @classmethod
    def create_indexes(cls):
        """ create indexes if they don't already exists """
        collection = cls.__collection__ or cls.__name__
        if not isinstance(cls.__database__, Database):
            raise NoDatabaseSet()
        if hasattr(cls, 'index_by') and type(cls.index_by) is list:
            for index in cls.index_by:
                if type(index) is dict and index.get('fields'):
                    options = index.get('options', {})
                    if not 'background' in options:
                        options['background'] = True
                    cls.__database__[collection].ensure_index(
                        index['fields'], **options)

    @property
    def id(self):
        """ id property returns mongodb's object _id """
        return self.get('_id')

    @classmethod
    def by_id(cls, id_value, *args, **kwargs):
        """ lookup record by id """
        if not isinstance(id_value, ObjectId):
            try:
                id_value = ObjectId(id_value)
            except InvalidId:
                pass
        return cls.find_one({'_id': id_value}, *args, **kwargs)

    @classmethod
    def by_slug(cls, slug, *args, **kwargs):
        """ lookup record by slug """
        return cls.find_one({'slug': slug}, *args, **kwargs)
